import unittest.mock
import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning)

import pytest
import httpx
import respx

import lightkube
from lightkube.config.kubeconfig import KubeConfig
from lightkube.resources.core_v1 import Pod, Node, Binding
from lightkube.generic_resource import create_global_resource
from lightkube.models.meta_v1 import ObjectMeta
from lightkube import types

from .test_client import make_wait_custom, make_wait_deleted, make_wait_failed, make_wait_success, make_watch_list, \
    json_contains

KUBECONFIG = """
apiVersion: v1
clusters:
- cluster: {server: 'https://localhost:9443'}
  name: test
contexts:
- context: {cluster: test, user: test}
  name: test
current-context: test
kind: Config
preferences: {}
users:
- name: test
  user: {token: testtoken}
"""


@pytest.fixture
def kubeconfig(tmpdir):
    kubeconfig = tmpdir.join("kubeconfig")
    kubeconfig.write(KUBECONFIG)
    return kubeconfig


@pytest.fixture
def kubeconfig_ns(tmpdir):
    kubeconfig = tmpdir.join("kubeconfig")
    kubeconfig.write(KUBECONFIG.replace('user: test', 'user: test, namespace: ns1'))
    return kubeconfig


@pytest.fixture
def client(kubeconfig):
    config = KubeConfig.from_file(str(kubeconfig))
    return lightkube.AsyncClient(config=config)


def test_namespace(client: lightkube.Client, kubeconfig_ns):
    assert client.namespace == 'default'

    config = KubeConfig.from_file(str(kubeconfig_ns))
    client = lightkube.AsyncClient(config=config)
    assert client.namespace == 'ns1'


@unittest.mock.patch('httpx.AsyncClient')
@unittest.mock.patch('lightkube.config.client_adapter.user_auth')
@unittest.mock.patch('lightkube.config.client_adapter.verify_cluster')
def test_client_httpx_attributes(verify_cluster, user_auth, httpx_async_client, kubeconfig):
    config = KubeConfig.from_file(kubeconfig)
    single_conf = config.get()
    lightkube.AsyncClient(config=single_conf, trust_env=False)
    verify_cluster.assert_called_once_with(single_conf.cluster, single_conf.user, single_conf.abs_file, trust_env=False)
    httpx_async_client.assert_called_once_with(
        timeout=None,
        base_url=single_conf.cluster.server,
        verify=verify_cluster.return_value,
        auth=user_auth.return_value,
        trust_env=False,
        transport=None,
    )


@respx.mock
@pytest.mark.asyncio
async def test_get_namespaced(client: lightkube.AsyncClient):
    respx.get("https://localhost:9443/api/v1/namespaces/default/pods/xx").respond(json={'metadata': {'name': 'xx'}})
    pod = await client.get(Pod, name="xx")
    assert pod.metadata.name == 'xx'

    respx.get("https://localhost:9443/api/v1/namespaces/other/pods/xx").respond(json={'metadata': {'name': 'xy'}})
    pod = await client.get(Pod, name="xx", namespace="other")
    assert pod.metadata.name == 'xy'
    await client.close()


@respx.mock
@pytest.mark.asyncio
async def test_list_global(client: lightkube.AsyncClient):
    resp = {'items': [{'metadata': {'name': 'xx'}}, {'metadata': {'name': 'yy'}}]}
    respx.get("https://localhost:9443/api/v1/nodes").respond(json=resp)
    nodes = client.list(Node)
    assert [node.metadata.name async for node in nodes] == ['xx', 'yy']

    respx.get("https://localhost:9443/api/v1/pods?fieldSelector=k%3Dx").respond(json=resp)
    pods = client.list(Pod, namespace=lightkube.ALL_NS, fields={'k': 'x'})
    assert [pod.metadata.name async for pod in pods] == ['xx', 'yy']

    # Binding doesn't support all namespaces
    with pytest.raises(ValueError):
        client.list(Binding, namespace=lightkube.ALL_NS)
    await client.close()


@respx.mock
@pytest.mark.asyncio
async def test_list_namespaced(client: lightkube.AsyncClient):
    resp = {'items':[{'metadata': {'name': 'xx'}}, {'metadata': {'name': 'yy'}}],
            'metadata': {'resourceVersion': '42'}}
    respx.get("https://localhost:9443/api/v1/namespaces/default/pods").respond(json=resp)
    poditer = client.list(Pod)
    with pytest.raises(lightkube.NotReadyError):
        poditer.resourceVersion
    pods = [pod async for pod in poditer]
    assert poditer.resourceVersion == "42"
    for pod, expected in zip(pods, resp["items"]):
        assert pod.metadata is not None
        assert pod.metadata.name == expected["metadata"]["name"]
        assert pod.apiVersion is not None
        assert pod.kind is not None
    
    respx.get("https://localhost:9443/api/v1/namespaces/other/pods?labelSelector=k%3Dv").respond(json=resp)
    pods = client.list(Pod, namespace="other", labels={'k': 'v'})
    assert [pod.metadata.name async for pod in pods] == ['xx', 'yy']


@respx.mock
@pytest.mark.asyncio
async def test_list_chunk_size(client: lightkube.AsyncClient):
    resp = {'items': [{'metadata': {'name': 'xx'}}, {'metadata': {'name': 'yy'}}], 'metadata': {'continue': 'yes'}}
    respx.get("https://localhost:9443/api/v1/namespaces/default/pods?limit=3").respond(json=resp)
    resp = {'items': [{'metadata': {'name': 'zz'}}]}
    respx.get("https://localhost:9443/api/v1/namespaces/default/pods?limit=3&continue=yes").respond(json=resp)
    pods = client.list(Pod, chunk_size=3)
    assert [pod.metadata.name async for pod in pods] == ['xx', 'yy', 'zz']
    await client.close()


@respx.mock
@pytest.mark.asyncio
async def test_delete_global(client: lightkube.AsyncClient):
    respx.delete("https://localhost:9443/api/v1/nodes/xx")
    await client.delete(Node, name="xx")

    # with cascade and grace_period
    respx.delete("https://localhost:9443/api/v1/nodes/params?propagationPolicy=Foreground&gracePeriodSeconds=0")
    await client.delete(Node, name="params", cascade=types.CascadeType.FOREGROUND, grace_period=0)

    # dry-run
    req_dry = respx.delete("https://localhost:9443/api/v1/nodes/xz?dryRun=All").respond(
        text="deleted")
    node = await client.delete(Node, name="xz", dry_run=True)
    assert req_dry.calls[0][0].url.params['dryRun'] == 'All'

    await client.close()

@respx.mock
@pytest.mark.asyncio
async def test_deletecollection_global(client: lightkube.AsyncClient):
    respx.delete("https://localhost:9443/api/v1/nodes")
    await client.deletecollection(Node)

    respx.delete("https://localhost:9443/api/v1/nodes?propagationPolicy=Foreground&gracePeriodSeconds=0")
    await client.deletecollection(Node, cascade=types.CascadeType.FOREGROUND, grace_period=0)

    await client.close()

@respx.mock
@pytest.mark.asyncio
async def test_deletecollection_namespaced(client: lightkube.AsyncClient):
    # dry-run
    req_dry = respx.delete("https://localhost:9443/api/v1/namespaces/other/pods?dryRun=All")
    pod = await client.deletecollection(Pod, namespace="other", dry_run=True)
    assert req_dry.calls[0][0].url.params['dryRun'] == 'All'
    await client.close()

@respx.mock
@pytest.mark.asyncio
async def test_watch(client: lightkube.AsyncClient):
    respx.get("https://localhost:9443/api/v1/nodes?watch=true").respond(content=make_watch_list())
    respx.get("https://localhost:9443/api/v1/nodes?watch=true&resourceVersion=1").respond(status_code=404)

    i = -1
    with pytest.raises(httpx.HTTPError) as exi:
        async for (op, node) in client.watch(Node):
            i += 1
            assert node.metadata.name == f'p{i}'
            assert op == 'ADDED'

    assert i == 9
    assert exi.value.response.status_code == 404
    await client.close()


@respx.mock
@pytest.mark.asyncio
async def test_watch_version(client: lightkube.AsyncClient):
    respx.get("https://localhost:9443/api/v1/nodes?resourceVersion=2&watch=true").respond(content=make_watch_list())
    respx.get("https://localhost:9443/api/v1/nodes?resourceVersion=1&watch=true").respond(status_code=404)

    # testing starting from specific resource version
    i = -1
    with pytest.raises(httpx.HTTPError) as exi:
        async for (op, node) in client.watch(Node, resource_version="2"):
            i += 1
            assert node.metadata.name == f'p{i}'
            assert op == 'ADDED'
    assert i == 9
    assert exi.value.response.status_code == 404
    await client.close()


@respx.mock
@pytest.mark.asyncio
async def test_wait_success(client: lightkube.AsyncClient):
    base_url = "https://localhost:9443/api/v1/nodes?fieldSelector=metadata.name%3Dtest-node&watch=true"

    respx.get(base_url).respond(content=make_wait_success())
    respx.get(base_url + "&resourceVersion=1").respond(content=make_wait_success())

    node = await client.wait(Node, "test-node", for_conditions=["TestCondition"])

    assert node.to_dict()["metadata"]["name"] == "test-node"

    await client.close()


@respx.mock
@pytest.mark.asyncio
async def test_wait_deleted(client: lightkube.AsyncClient):
    base_url = "https://localhost:9443/api/v1/nodes?fieldSelector=metadata.name%3Dtest-node&watch=true"

    respx.get(base_url).respond(content=make_wait_deleted())
    respx.get(base_url + "&resourceVersion=1").respond(content=make_wait_deleted())

    message = "nodes/test-node was unexpectedly deleted"
    with pytest.raises(lightkube.core.exceptions.ObjectDeleted, match=message):
        await client.wait(Node, "test-node", for_conditions=["TestCondition"])

    await client.close()


@respx.mock
@pytest.mark.asyncio
async def test_wait_failed(client: lightkube.AsyncClient):
    base_url = "https://localhost:9443/api/v1/nodes?fieldSelector=metadata.name%3Dtest-node&watch=true"

    respx.get(base_url).respond(content=make_wait_failed())
    respx.get(base_url + "&resourceVersion=1").respond(content=make_wait_failed())

    message = r"nodes/test-node has failure condition\(s\): TestCondition"
    with pytest.raises(lightkube.core.exceptions.ConditionError, match=message):
        await client.wait(Node, "test-node", for_conditions=[], raise_for_conditions=["TestCondition"])

    await client.close()


@respx.mock
@pytest.mark.asyncio
async def test_wait_custom(client: lightkube.AsyncClient):
    base_url = "https://localhost:9443/apis/custom.org/v1/customs?fieldSelector=metadata.name%3Dcustom-resource&watch=true"

    Custom = create_global_resource(
        group="custom.org", version="v1", kind="Custom", plural="customs"
    )
    respx.get(base_url).respond(content=make_wait_custom())
    respx.get(base_url + "&resourceVersion=1").respond(content=make_wait_custom())

    await client.wait(Custom, "custom-resource", for_conditions=["TestCondition"])

    await client.close()


@respx.mock
@pytest.mark.asyncio
async def test_patch_global(client: lightkube.AsyncClient):
    req = respx.patch("https://localhost:9443/api/v1/nodes/xx").respond(json={'metadata': {'name': 'xx'}})
    pod = await client.patch(Node, "xx", [{"op": "add", "path": "/metadata/labels/x", "value": "y"}],
                             patch_type=types.PatchType.JSON)
    assert pod.metadata.name == 'xx'
    assert req.calls[0][0].headers['Content-Type'] == "application/json-patch+json"

    # PatchType.APPLY + force
    req = respx.patch("https://localhost:9443/api/v1/nodes/xy?fieldManager=test&force=true").respond(
        json={'metadata': {'name': 'xy'}})
    node = await client.patch(Node, "xy", Pod(metadata=ObjectMeta(labels={'l': 'ok'})),
                              patch_type=types.PatchType.APPLY, field_manager='test', force=True)
    assert node.metadata.name == 'xy'
    assert req.calls[0][0].headers['Content-Type'] == "application/apply-patch+yaml"

    # dry-run
    req = respx.patch("https://localhost:9443/api/v1/nodes/xz?fieldManager=test&dryRun=All").respond(
        json={'metadata': {'name': 'xz'}})
    node = await client.patch(Node, "xz", [{"op": "add", "path": "/metadata/labels/x", "value": "z"}],
                            patch_type=types.PatchType.JSON, field_manager='test', dry_run=True)
    assert node.metadata.name == 'xz'
    assert req.calls[0][0].url.params['dryRun'] == 'All'

    await client.close()


@respx.mock
@pytest.mark.asyncio
async def test_create_global(client: lightkube.AsyncClient):
    req = respx.post("https://localhost:9443/api/v1/nodes").respond(json={'metadata': {'name': 'xx'}})
    pod = await client.create(Node(metadata=ObjectMeta(name="xx")))
    json_contains(req.calls[0][0].read(), {"metadata": {"name": "xx"}})
    assert pod.metadata.name == 'xx'

    # dry-run
    req_dry = respx.post("https://localhost:9443/api/v1/nodes").respond(
        json={'metadata': {'name': 'xx'}})
    node = await client.create(Node(metadata=ObjectMeta(name='xx')), dry_run=True)
    assert req_dry.calls[1][0].url.params['dryRun'] == 'All'

    await client.close()


@respx.mock
@pytest.mark.asyncio
async def test_replace_global(client: lightkube.AsyncClient):
    req = respx.put("https://localhost:9443/api/v1/nodes/xx").respond(json={'metadata': {'name': 'xx'}})
    pod = await client.replace(Node(metadata=ObjectMeta(name="xx")))
    json_contains(req.calls[0][0].read() , {"metadata": {"name": "xx"}})
    assert pod.metadata.name == 'xx'

    # dry-run
    req_dry = respx.put("https://localhost:9443/api/v1/nodes/xx").respond(
        json={'metadata': {'name': 'xx'}})
    pod = await client.replace(Node(metadata=ObjectMeta(name='xx')), dry_run=True)
    assert req_dry.calls[1][0].url.params['dryRun'] == 'All'

    await client.close()


async def alist(aiter):
    return [item async for item in aiter]


@respx.mock
@pytest.mark.asyncio
async def test_pod_log(client: lightkube.AsyncClient):
    result = ['line1\n', 'line2\n', 'line3\n']
    content = "".join(result)

    respx.get("https://localhost:9443/api/v1/namespaces/default/pods/test/log").respond(content=content)
    lines = await alist(client.log('test'))
    assert lines == result

    respx.get("https://localhost:9443/api/v1/namespaces/default/pods/test/log?since=30&timestamps=true").respond(
        content=content)
    lines = await alist(client.log('test', since=30, timestamps=True))
    assert lines == result

    respx.get("https://localhost:9443/api/v1/namespaces/default/pods/test/log?container=bla").respond(
        content=content)

    lines = await alist(client.log('test', container="bla", newlines=False))
    assert lines == [_.strip() for _ in result]

    lines = await alist(client.log('test', container="bla"))
    assert lines == result

    await client.close()

@respx.mock
@pytest.mark.asyncio
async def test_apply_namespaced(client: lightkube.AsyncClient):
    req = respx.patch("https://localhost:9443/api/v1/namespaces/default/pods/xy?fieldManager=test").respond(
        json={'metadata': {'name': 'xy'}})
    pod = await client.apply(Pod(metadata=ObjectMeta(name='xy')), field_manager='test')
    assert pod.metadata.name == 'xy'
    assert req.calls[0][0].headers['Content-Type'] == "application/apply-patch+yaml"

    # custom namespace, force
    req = respx.patch("https://localhost:9443/api/v1/namespaces/other/pods/xz?fieldManager=a&force=true").respond(
        json={'metadata': {'name': 'xz'}})
    pod = await client.apply(Pod(metadata=ObjectMeta(name='xz', namespace='other')), field_manager='a', force=True)
    assert pod.metadata.name == 'xz'
    assert req.calls[0][0].headers['Content-Type'] == "application/apply-patch+yaml"

    # sub-resource
    req = respx.patch("https://localhost:9443/api/v1/namespaces/default/pods/xx/status?fieldManager=a").respond(
        json={'metadata': {'name': 'xx'}})
    pod = await client.apply(Pod.Status(), name='xx', field_manager='a')
    assert pod.metadata.name == 'xx'
    assert req.calls[0][0].headers['Content-Type'] == "application/apply-patch+yaml"
    await client.close()


@respx.mock
@pytest.mark.asyncio
async def test_apply_global(client: lightkube.AsyncClient):
    req = respx.patch("https://localhost:9443/api/v1/nodes/xy?fieldManager=test").respond(
        json={'metadata': {'name': 'xy'}})
    node = await client.apply(Node(metadata=ObjectMeta(name='xy')), field_manager='test')
    assert node.metadata.name == 'xy'
    assert req.calls[0][0].headers['Content-Type'] == "application/apply-patch+yaml"

    # dry-run
    req = respx.patch("https://localhost:9443/api/v1/nodes/xy?fieldManager=test&dryRun=All").respond(
        json={'metadata': {'name': 'xy'}})
    node = await client.apply(Node(metadata=ObjectMeta(name='xy')), field_manager='test', dry_run=True)
    assert node.metadata.name == 'xy'
    assert req.calls[0][0].url.params['dryRun'] == 'All'

    # sub-resource + force
    req = respx.patch("https://localhost:9443/api/v1/nodes/xx/status?fieldManager=a&force=true").respond(
        json={'metadata': {'name': 'xx'}})
    node = await client.apply(Node.Status(), name='xx', field_manager='a', force=True)
    assert node.metadata.name == 'xx'
    assert req.calls[0][0].headers['Content-Type'] == "application/apply-patch+yaml"
    await client.close()



import decimal
import pytest

from lightkube.models.core_v1 import ResourceRequirements
from lightkube.utils.quantity import parse_quantity
from lightkube.utils.quantity import equals_canonically


def test_unitless():
    """Unitless values must be interpreted as decimal notation."""
    assert parse_quantity("1.5") == decimal.Decimal("1.5")
    assert parse_quantity("-1.5") == decimal.Decimal("-1.5")
    assert parse_quantity("0.30000000000000004") == decimal.Decimal("0.301")
    assert parse_quantity("0.09999999999999998") == decimal.Decimal("0.1")
    assert parse_quantity("3.141592653") == decimal.Decimal("3.142")


def test_binary_notation():
    assert parse_quantity("1.5Gi") == parse_quantity("1536Mi") == decimal.Decimal("1610612736")
    assert parse_quantity("0.9Gi") == decimal.Decimal("966367641.6")


def test_decimal_notation():
    assert parse_quantity("1.5G") == decimal.Decimal("1500000000")
    assert parse_quantity("0.9G") == decimal.Decimal("900000000")
    assert parse_quantity("500m") == decimal.Decimal("0.5")


def test_none():
    assert parse_quantity(None) is None


def test_invalid_value():
    with pytest.raises(ValueError):
        parse_quantity("1.2.3")
    with pytest.raises(ValueError):
        parse_quantity("1e2.3")
    with pytest.raises(ValueError):
        # decimal.InvalidOperation
        parse_quantity("9e999")
    with pytest.raises(ValueError):
        # decimal.Overflow
        parse_quantity("9e9999999")


def test_invalid_unit():
    with pytest.raises(ValueError):
        parse_quantity("1kb")
    with pytest.raises(ValueError):
        parse_quantity("1GGi")


def test_whitespace():
    with pytest.raises(ValueError):
        parse_quantity("")
    with pytest.raises(ValueError):
        parse_quantity(" ")
    with pytest.raises(ValueError):
        parse_quantity("1 ")
    with pytest.raises(ValueError):
        parse_quantity(" 1")
    with pytest.raises(ValueError):
        parse_quantity("1 Gi")


def test_canonical_equality_for_dicts_with_blanks():
    first = {}
    second = {}
    assert equals_canonically(first, second)

    first = {}
    second = None
    assert equals_canonically(first, second)


def test_canonical_equality_for_dicts_with_cpu():
    first = {"cpu": "0.5"}
    second = {"cpu": "500m"}
    assert equals_canonically(first, second)


def test_canonical_equality_for_dicts_with_memory():
    first = {"memory": "1G"}
    second = {"memory": "1Gi"}
    assert not equals_canonically(first, second)


def test_canonical_equality_for_dicts_with_both():
    first = {"cpu": "0.6", "memory": "1.5Gi"}
    second = {"cpu": "600m", "memory": "1536Mi"}
    assert equals_canonically(first, second)


def test_canonical_equality_for_extended_resources():
    first = {"cpu": "0.6", "example.com/foo": "1"}
    second = {"cpu": "600m", "example.com/foo": "1"}
    assert equals_canonically(first, second)

    first = {"cpu": "0.6", "example.com/foo": "1"}
    second = {"cpu": "600m", "example.com/foo": "2"}
    assert not equals_canonically(first, second)

    first = {"cpu": "0.6", "example.com/foo": "1"}
    second = {"cpu": "600m"}
    assert not equals_canonically(first, second)


def test_canonical_equality_for_resource_requirements_with_blanks():
    first = ResourceRequirements()
    second = ResourceRequirements()
    assert equals_canonically(first, second)

    first = ResourceRequirements(limits={})
    second = ResourceRequirements(limits={})
    assert equals_canonically(first, second)

    first = ResourceRequirements(limits={})
    second = ResourceRequirements(requests={})
    assert equals_canonically(first, second)


def test_canonical_equality_for_resource_requirements_with_cpu():
    first = ResourceRequirements(limits={"cpu": "0.5"})
    second = ResourceRequirements(limits={"cpu": "500m"})
    assert equals_canonically(first, second)

    first = ResourceRequirements(requests={"cpu": "0.5"})
    second = ResourceRequirements(requests={"cpu": "500m"})
    assert equals_canonically(first, second)

    first = ResourceRequirements(limits={"cpu": "0.5"})
    second = ResourceRequirements(requests={"cpu": "500m"})
    assert not equals_canonically(first, second)

    first = ResourceRequirements(limits={"cpu": "0.6"}, requests={"cpu": "0.5"})
    second = ResourceRequirements(limits={"cpu": "600m"}, requests={"cpu": "500m"})
    assert equals_canonically(first, second)


def test_canonical_equality_for_resource_requirements_with_memory():
    first = ResourceRequirements(limits={"memory": "1G"})
    second = ResourceRequirements(limits={"memory": "1Gi"})
    assert not equals_canonically(first, second)


def test_canonical_equality_for_resource_requirements_with_both():
    first = ResourceRequirements(limits={"cpu": "0.6", "memory": "1.5Gi"}, requests={"cpu": "0.5"})
    second = ResourceRequirements(limits={"cpu": "600m", "memory": "1536Mi"}, requests={"cpu": "500m"})
    assert equals_canonically(first, second)


def test_invalid_canonical_equality():
    with pytest.raises(TypeError):
        equals_canonically({}, ResourceRequirements())
    with pytest.raises(TypeError):
        equals_canonically(None, ResourceRequirements())



from lightkube import operators


def test_exists():
    assert operators.exists().encode('key') == 'key'


def test_not_exists():
    assert operators.not_exists().encode('key') == '!key'


def test_equal():
    assert operators.equal('xxx').encode('key') == 'key=xxx'


def test_not_equal():
    assert operators.not_equal('xxx').encode('key') == 'key!=xxx'


def test_in():
    assert operators.in_(['xxx', 'yyy']).encode('key') == 'key in (xxx,yyy)'


def test_not_in():
    assert operators.not_in(['xxx', 'zzz']).encode('key') == 'key notin (xxx,zzz)'



import pytest

from lightkube.core.selector import build_selector
from lightkube import operators


def test_simple_types():
    r = build_selector({
        'k1': 'v1',
        'k2': None,
        'k3': ['b', 'c'],
        'k4': {'a', 'b'},
        'k5': ('d', 'b')
    })

    assert r == "k1=v1,k2,k3 in (b,c),k4 in (a,b),k5 in (b,d)"


def test_operators():
    r = build_selector({
        'k1': operators.equal('v1'),
        'k2': operators.not_exists(),
        'k3': operators.in_(['b', 'c']),
        'k4': operators.not_in(['b', 'c']),
        'k5': operators.not_equal('v5'),
        'k6': operators.exists()
    })

    assert r == "k1=v1,!k2,k3 in (b,c),k4 notin (b,c),k5!=v5,k6"


def test_binary_only_selector():
    with pytest.raises(ValueError):
        build_selector({'k2': None}, for_fields=True)

    with pytest.raises(ValueError):
        build_selector({'k2': operators.in_(['b', 'c'])}, for_fields=True)

    r = build_selector({'k1': 'a', 'k2': operators.not_equal('a')}, for_fields=True)
    assert r == "k1=a,k2!=a"

    r = build_selector({'k1': 'a', 'k2': operators.not_in(['a', 'b'])}, for_fields=True)
    assert r == "k1=a,k2!=a,k2!=b"



import importlib
from typing import Optional, Type, TypeVar, Union

from lightkube.core import resource as res
from lightkube.core.exceptions import LoadResourceError

AnyResource = Union[res.NamespacedResource, res.GlobalResource]
AnyResourceType = Type[AnyResource]
AnyResourceTypeVar = TypeVar("AnyResourceTypeVar", bound=AnyResourceType)


def _load_internal_resource(version, kind):
    if "/" in version:
        group, version_n = version.split("/")
        # Generic resource not defined, but it could be a k8s resource
        if group.endswith(".k8s.io"):
            group = group[:-7]
        group = group.replace(".", "_")
        module_name = "_".join([group, version_n])
    else:
        module_name = f"core_{version}"

    module = importlib.import_module(f"lightkube.resources.{module_name.lower()}")
    try:
        return getattr(module, kind)
    except AttributeError:
        raise LoadResourceError(
            f"Cannot find resource kind '{kind}' in module {module.__name__}"
        )


def _maybe_internal(version):
    if "/" not in version:
        return True

    group = version.split("/")[0]
    # internal resources don't have namespace or end in .k8s.io
    return group.endswith(".k8s.io") or "." not in group


class ResourceRegistry:
    """Resource Registry used to load standard resources or to register custom resources"""

    _registry: dict

    def __init__(self):
        self._registry = {}

    def register(self, resource: AnyResourceTypeVar) -> AnyResourceTypeVar:
        """Register a custom resource

        **parameters**

        * **resource** - Resource class to register.

        **returns** The `resource` class provided
        """
        info = resource._api_info
        version = (
            f"{info.resource.group}/{info.resource.version}"
            if info.resource.group
            else info.resource.version
        )
        res_key = (version, info.resource.kind)

        if res_key in self._registry:
            registered_resource = self._registry[res_key]
            if registered_resource is resource:  # already present
                return registered_resource
            raise ValueError(
                f"Another class for resource '{info.resource.kind}' is already registered"
            )

        self._registry[res_key] = resource
        return resource

    def clear(self):
        """Clear the registry from all registered resources"""
        self._registry.clear()

    def get(self, version: str, kind: str) -> Optional[Type[AnyResource]]:
        """Get a resource from the registry matching the given `version` and `kind`.

        **parameters**

        * **version** - Version of the resource as defined in the kubernetes definition. Example `example.com/v1`
        * **kind** - Resource kind. Example `CronJob`

        **returns** A `resource` class or `None` if there is no match in the registry.
        """
        return self._registry.get((version, kind))

    def load(self, version, kind) -> Optional[Type[AnyResource]]:
        """Load a standard resource from `lightkube.resources` given `version` and `kind`.
        This method look up the registry first and import the resource from the module only if it's not available there.

        * **version** - Version of the resource as defined in the kubernetes definition. Example `apps/v1`
        * **kind** - Resource kind. Example `Pod`

        **returns** A `resource` class if the resource is found. Otherwise an exception is raised
        """
        # check if this resource is in the registry
        resource = self.get(version, kind)
        if resource is not None:
            return resource

        # if not, we attempt to load from lightkube.resources
        if _maybe_internal(version):
            try:
                return self.register(_load_internal_resource(version, kind))
            except ImportError:
                pass

        raise LoadResourceError(
            f"Cannot find resource {kind} of group {version}. "
            "If using a CRD, ensure a generic resource is defined."
        )


resource_registry = ResourceRegistry()



"""
This module exposes dependencies used by lightkube-models

These dependencies are here because we may decide to replace dataclasses with something else in the future
"""
__all__ = ["dataclass", "field", "DictMixin"]
from dataclasses import dataclass, field
from .dataclasses_dict import DataclassDictMixIn as DictMixin



from collections import namedtuple
import pytest

from lightkube import sort_objects


@pytest.fixture()
def resources_in_apply_order():
    mock_resource = namedtuple("resource", ("kind",))
    resources = [
        mock_resource(kind="CustomResourceDefinition"),
        mock_resource(kind="Namespace"),
        mock_resource(kind="Secret"),
        mock_resource(kind="ServiceAccount"),
        mock_resource(kind="PersistentVolume"),
        mock_resource(kind="PersistentVolumeClaim"),
        mock_resource(kind="ConfigMap"),
        mock_resource(kind="Role"),
        mock_resource(kind="ClusterRole"),
        mock_resource(kind="RoleBinding"),
        mock_resource(kind="ClusterRoleBinding"),
        mock_resource(kind="something-else"),
    ]
    return resources


@pytest.mark.parametrize(
    "reverse",
    [
        False,  # Desired result in apply-friendly order
        True,   # Desired order in delete-friendly order
    ]
)
def test_sort_objects_by_kind(reverse, resources_in_apply_order):
    """Tests that sort_objects can kind-sort objects in both apply and delete orders."""
    resources_expected_order = resources_in_apply_order
    if reverse:
        resources_expected_order = list(reversed(resources_expected_order))

    # Add disorder to the test input
    resources_unordered = resources_expected_order[1:] + [resources_expected_order[0]]

    result = sort_objects(resources_unordered, reverse=reverse)
    assert result == resources_expected_order



from typing import Iterable

__all__ = ["in_", "not_in", "exists", "not_exists", "equal", "not_equal"]


class Operator:
    def __init__(self, op_name: str, op: str, value=None):
        self.op = op
        self.value = value
        self.op_name = op_name

    def encode(self, key):
        return f"{key}{self.op}{self.value}"


class SequenceOperator(Operator):
    def encode(self, key):
        return f"{key} {self.op} ({','.join(self.value)})"


class BinaryOperator(Operator):
    pass


class UnaryOperator(Operator):
    def encode(self, key):
        return f"{self.op}{key}"


def in_(values: Iterable) -> SequenceOperator:
    return SequenceOperator("in_", "in", sorted(values))


def not_in(values: Iterable) -> SequenceOperator:
    return SequenceOperator("not_in", "notin", sorted(values))


def exists() -> UnaryOperator:
    return UnaryOperator("exists", "")


def not_exists() -> UnaryOperator:
    return UnaryOperator("not_exists", "!")


def equal(value: str) -> BinaryOperator:
    return BinaryOperator("equal", "=", value)


def not_equal(value: str) -> BinaryOperator:
    return BinaryOperator("not_equal", "!=", value)
